const std = @import("std");

pub fn Slot(comptime T: type) type {
    return struct {
        value: T,
        version: std.atomic.Value(usize),
    };
}

pub fn RingBuffer(comptime T: type) type {
    return struct {
        const Self = @This();

        capacity: usize,
        buffer: []Slot(T),
        head: std.atomic.Value(usize) = std.atomic.Value(usize).init(0),
        tail: usize = 0,

        pub fn init(buffer: []Slot(T)) error{CapacityTooSmall}!Self {
            if (buffer.len <= 1) {
                return error.CapacityTooSmall;
            }

            for (buffer, 0..) |*slot, i| {
                slot.version = std.atomic.Value(usize).init(i);
            }

            return .{
                .capacity = buffer.len,
                .buffer = buffer,
            };
        }

        pub fn enqueue(self: *Self, value: T) error{Overflow}!void {
            while (true) {
                const head = self.head.load(.acquire);
                const index = head % self.capacity;
                const slot = &self.buffer[index];

                const expected_version = head;
                if (slot.version.load(.acquire) != expected_version) {
                    return error.Overflow;
                }

                if (self.head.cmpxchgStrong(
                    head,
                    head + 1,
                    .seq_cst,
                    .seq_cst,
                )) |_| {
                    std.atomic.spinLoopHint();
                    continue;
                }

                slot.value = value;
                slot.version.store(expected_version + 1, .release);
                return;
            }
        }

        pub fn dequeue(self: *Self) error{Underflow}!T {
            const tail = self.tail;
            const index = tail % self.capacity;
            const slot = &self.buffer[index];

            const expected_version = tail + 1;
            if (slot.version.load(.acquire) != expected_version) {
                return error.Underflow;
            }

            const value = slot.value;
            slot.version.store(tail +% self.capacity, .release);
            self.tail +%= 1;
            return value;
        }

        pub fn dequeueAll(self: *Self, buf: []T) ![]T {
            if (buf.len < self.capacity) {
                return error.BufferTooSmall;
            }

            const head = self.head.load(.acquire) % self.capacity;
            var tail = self.tail % self.capacity;
            var i: usize = 0;

            while (true) {
                const val = self.dequeue() catch break;
                buf[i] = val;
                i += 1;
                tail = (tail + 1) % self.capacity;
                if (tail == head) break;
            }

            return buf[0..i];
        }
    };
}
